from pynir.utils import simulateNIR_calibrationTransfer
from pynir.Calibration import pls, regresssionReport
from pynir.CalibrationTransfer import PDS,SST, BS
import matplotlib.pyplot as plt
import numpy as np

from sklearn.model_selection import train_test_split

# Simulate NIR spectra for calibration transfer
nSample = 100
X1, X2, y, wv = simulateNIR_calibrationTransfer(nSample=nSample,nComp=10,shifts=5e1)
idxTrain,idxTest = train_test_split(np.arange(nSample),test_size=0.6)
idxTransfer,idxTest = train_test_split(idxTest,test_size=0.5)


fig, ax = plt.subplots(2,sharex=True)
ax[0].plot(wv, np.transpose(X1))
ax[0].set_title("First")

ax[1].plot(wv, np.transpose(X2))
ax[1].set_title("Second")

fig.supxlabel("Wavelength (nm)")
fig.supylabel("Intensity(A.U.)")
plt.show()

# Multivariate calibration
nComp = 7
plsModel1 = pls(nComp=nComp).fit(X1[idxTrain,:], y[idxTrain])

yhat1 = plsModel1.predict(X1[idxTest,:],nComp=nComp)
yhat2= plsModel1.predict(X2[idxTest,:],nComp=nComp)

fig, ax = plt.subplots(2,sharex=True,figsize=(8,16))
plsModel1.plot_prediction(y[idxTest], yhat1,title = "First", ax = ax[0])
plsModel1.plot_prediction(y[idxTest], yhat2, title= "Second", ax = ax[1])

# Calibration transfer on spectra
## PDS
X2_PDS = PDS(halfWindowSize=3).fit(X1[idxTransfer,:], X2[idxTransfer,:]).transform(X2[idxTest,:])
yhat2_PDS= plsModel1.predict(X2_PDS,nComp=nComp)
plsModel1.plot_prediction(y[idxTest], yhat2_PDS, title= "PDS")

fig, ax = plt.subplots()
ax.plot(wv, np.transpose(X2_PDS))


## SST
X2_SST = SST(nComp=nComp).fit(X1[idxTransfer,:], X2[idxTransfer,:]).transform(X2[idxTest,:])
yhat2_SST= plsModel1.predict(X2_SST,nComp=nComp)
plsModel1.plot_prediction(y[idxTest], yhat2_SST, title= "SST")

fig, ax = plt.subplots()
ax.plot(wv, np.transpose(X2_SST))


# Calibration transfer on prediction
## BS
yhat2_BS = BS().fit(yhat1, yhat2).transform(yhat2)
plsModel1.plot_prediction(y[idxTest], yhat2_BS, title= "BS")


import numpy as np
import matplotlib.pyplot as plt

from pynir.utils import simulateNIR
from pynir.Calibration import pls
from pynir.Calibration import regresssionReport

from sklearn.model_selection import train_test_split

# simulate NIR data
X,y,wv = simulateNIR(nSample=200,nComp=10,noise=1e-5)

Xtrain, Xtest, ytrain,ytest = train_test_split(X,y,test_size=0.2)

# estabilish PLS model
nComp = 10
plsModel = pls(nComp = nComp)
plsModel.fit(Xtrain,ytrain)

# 10 fold cross validation for selecting optimal nComp
yhat_train = plsModel.predict(Xtrain,nComp=np.arange(nComp)+1)
yhat_cv    = plsModel.crossValidation_predict(nfold = 10)
yhat_test  = plsModel.predict(Xtest,nComp=np.arange(nComp)+1)

rmsec = []
rmsecv = []
rmsep = []

r2c = []
r2cv = []
r2p = []

for i in range(nComp):
    report_train = regresssionReport(ytrain, yhat_train[:,i])
    report_cv = regresssionReport(ytrain, yhat_cv[:,i])
    report_test = regresssionReport(ytest, yhat_test[:,i])
    
    rmsec.append(report_train["rmse"])
    rmsecv.append(report_cv["rmse"])
    rmsep.append(report_test["rmse"])
    
    r2c.append(report_train["r2"])
    r2cv.append(report_cv["r2"])
    r2p.append(report_test["r2"])
    
fig,ax = plt.subplots()
ax.plot(np.arange(nComp)+1,rmsec, marker = ".",label = "RMSEC")
ax.plot(np.arange(nComp)+1,rmsecv, marker = "*",label = "RMSECV")
ax.plot(np.arange(nComp)+1,rmsep, marker = "^",label = "RMSEP")
ax.set_xlabel("nComp")
ax.set_ylabel("RMSE")
ax.legend()
plt.show()

optLV = plsModel.get_optLV()  # optimized nComp based on cross validation
yhat_test_opt = plsModel.predict(Xtest, nComp = optLV)

fig, ax = plt.subplots()
ax.plot(ytest, yhat_test_opt, "*", label = "Optimal prediction of test set")
ax.plot([np.min(y), np.max(y)], [np.min(y), np.max(y)], label = "y = x")
ax.set_xlabel("Reference")
ax.set_ylabel("Prediction")
ax.legend()
plt.show()


fig, ax = plt.subplots()
ax.plot(wv,plsModel.model["B"][1:,optLV-1], label = "PLS regresssion coefficients")
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Coefficients")
ax.legend()
plt.show()




# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 11:00:35 2022

@author: chinn
"""
import numpy as np
from sklearn.cross_decomposition import PLSRegression
from scipy.stats import f, chi2

import matplotlib.pyplot as plt

class outlierDection_PLS():
    # Ref1: https://nirpyresearch.com/outliers-detection-pls-regression-nir-spectroscopy-python/
    # Ref2: https://www.sciencedirect.com/science/article/pii/S0378517314004980
    def __init__(self, ncomp = 2, conf = 0.99):
        self.ncomp = ncomp
        self.conf = conf
    
    def fit(self, X, y):
        ncomp = self.ncomp
        self.plsModel = PLSRegression(n_components=ncomp)
        self.plsModel.fit(X, y)
        
        return self
    
    def detect(self, X, y):
        ncomp = self.ncomp
        conf = self.conf
        plsModel = self.plsModel
        # Get X scores
        T = plsModel.transform(X)

        # Calculate error array
        Err = X - plsModel.inverse_transform(T)
         
        # Calculate Q-residuals (sum over the rows of the error array)
        Q = np.sum(Err**2, axis=1)
        
        # Estimate the confidence level for the Q-residuals
        Q_conf = (np.var(Q)/2/np.mean(Q))*chi2.ppf(conf, 2*np.mean(Q)**2/np.var(Q))
        
        # Calculate Hotelling's T-squared (note that data are normalised by default)
        Tsq = np.sum((plsModel.x_scores_/np.std(plsModel.x_scores_, axis=0))**2, axis=1)
        
        # Calculate confidence level for T-squared from the ppf of the F distribution
        Tsq_conf = f.ppf(q=conf,dfn=ncomp,dfd=X.shape[0]-ncomp)
        Tsq_conf = Tsq_conf*ncomp*(X.shape[0]-1)*(X.shape[0]+1)/X.shape[0]/(X.shape[0]-ncomp)
        idxOutlier = np.logical_and(Q>Q_conf, Tsq>Tsq_conf)
         
        return Q, Tsq, Q_conf, Tsq_conf, idxOutlier
    
    def plot_HotellingT2_Q(self, Q, Tsq, Q_conf, Tsq_conf, ax=None):
        if ax ==None:
            fig, ax = plt.subplots(figsize=(8,4.5))
    
        ax.plot(Tsq, Q, 'o')
     
        ax.plot([Tsq_conf,Tsq_conf],[plt.axis()[2],plt.axis()[3]],  '--')
        ax.plot([plt.axis()[0],plt.axis()[1]],[Q_conf,Q_conf],  '--')
        ax.set_xlabel("Hotelling's T-squared")
        ax.set_ylabel('Q residuals')
     
        plt.show()


import os
import sys

current_dir = os.path.dirname(__file__)
parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
pynir_dir = os.path.join(parent_dir, 'src')
if pynir_dir not in sys.path:
    sys.path.insert(0, pynir_dir)


import numpy as np
import matplotlib.pyplot as plt

from pynir.utils import simulateNIR
from pynir.Calibration import plsda
from pynir.Calibration import binaryClassificationReport,plot_confusion_matrix

from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix


# simulate NIR data
X,y,wv = simulateNIR(nSample=200,nComp=10,refType=2, noise=1e-5)

Xtrain, Xtest, ytrain,ytest = train_test_split(X,y,test_size=0.2)

# estabilish PLS model
nComp = 10
plsdaModel = plsda(n_components = nComp).fit(Xtrain,ytrain)

# 10 fold cross validation for selecting optimal nlv
accuracy_cv = []
yhat_cv  = plsdaModel.crossValidation_predict(nfold = 10)
for i in range(nComp):
    report_cv = binaryClassificationReport(ytrain, yhat_cv[:,i])
    accuracy_cv.append(report_cv["accuracy"])


fig,ax = plt.subplots()
ax.plot(np.arange(nComp)+1,accuracy_cv, marker = "*",label = "Accuracy$_c$$_v$")
ax.set_xlabel("nLV")
ax.set_ylabel("Accuracy")
ax.legend()
plt.show()


optLV = plsdaModel.get_optLV()  # optimized nLV based on cross validation

plsdaModel_opt = plsda(n_components = optLV)
plsdaModel_opt.fit(Xtrain,ytrain)
yhat_train_opt = plsdaModel_opt.predict(Xtrain)
yhat_test_opt = plsdaModel_opt.predict(Xtest)

cm_train_opt = confusion_matrix(y_true=ytrain, y_pred=yhat_train_opt)
cm_test_opt = confusion_matrix(y_true=ytest, y_pred=yhat_test_opt)



plot_confusion_matrix(cm_train_opt,np.unique(y),normalize=False,
                      title="Confusion matrix for prediction on training set")
plot_confusion_matrix(cm_test_opt,np.unique(y),normalize=False,
                      title="Confusion matrix for prediction on testing set")

#vip =plsdaModel_opt.get_vip()
#fig,ax = plt.subplots()
#ax.plot(vip)

from sklearn.model_selection import cross_val_predict
from sklearn.metrics import r2_score

q2, r2, permutation_ratios = plsdaModel_opt.permutation_test(X, y, n_repeats = 1000, n_jobs=-1)

# Calculate the observed Q2 and R2 values
y_pred = cross_val_predict(plsdaModel_opt, X, y, cv=10)
q2_observed = plsdaModel_opt.score(X, y)
r2_observed = r2_score(y, y_pred)

permutation_ratios = np.append(permutation_ratios,0)
q2 = np.append(q2,q2_observed)
r2 = np.append(r2,r2_observed)

# Fit a linear regression model to the permutation test data
q2_coef, q2_intercept = np.polyfit(permutation_ratios, q2, 1)
r2_coef, r2_intercept = np.polyfit(permutation_ratios, r2, 1)

# Calculate the fitted line values
q2_fit = q2_coef * permutation_ratios + q2_intercept
r2_fit = r2_coef * permutation_ratios + r2_intercept

# Plot the fitted lines
plt.scatter(1-permutation_ratios, q2, label='Q2')
plt.scatter(1-permutation_ratios, r2, label='R2')
plt.plot(1-permutation_ratios, q2_fit, '--', color='black', label='Q2 Fit')
plt.plot(1-permutation_ratios, r2_fit, '--', color='gray', label='R2 Fit')


plt.xlabel('Retain Ratio')
plt.ylabel('Model Performance')
plt.legend()
plt.show()



import numpy as np
from sklearn.cross_decomposition import PLSRegression
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import ShuffleSplit
from sklearn.model_selection import cross_val_predict
from sklearn.model_selection import cross_val_score
from sklearn.utils import shuffle
from numpy.linalg import matrix_rank as rank

import matplotlib.pyplot as plt
from .Preprocessing import smooth


# Single step feature selection method
class MCUVE:
    def __init__(self, x, y, ncomp=2, nrep=500, testSize=0.2):
        self.x = x
        self.y = y
        # The number of latent components should not be larger than any dimension size of independent matrix
        self.ncomp = min([ncomp, rank(x)])
        self.nrep = nrep
        self.testSize = testSize


    def fit(self):
        PLSCoef = np.zeros((self.nrep, self.x.shape[1]))
        ss = ShuffleSplit(n_splits=self.nrep, test_size=self.testSize)
        step = 0
        for train, test in ss.split(self.x, self.y):
            xtrain = self.x[train, :]
            ytrain = self.y[train]
            plsModel = PLSRegression(min([self.ncomp, rank(xtrain)]))
            plsModel.fit(xtrain, ytrain)
            PLSCoef[step, :] = plsModel.coef_.T
            step += 1
        meanCoef = np.mean(PLSCoef, axis=0)
        stdCoef = np.std(PLSCoef, axis=0)
        self.criteria = meanCoef / stdCoef
        self.featureRank = np.argsort(-np.abs(self.criteria))
        return self
    
    def evalFeatures(self, cv=5):
        self.featureR2 = []
        for i in range(self.x.shape[1]):
            xi = self.x[:, self.featureRank[:i + 1]]
            if i<self.ncomp:
                regModel = LinearRegression()
            else:
                regModel = PLSRegression(min([self.ncomp, rank(xi)]))

            cvScore = cross_val_score(regModel, xi, self.y, cv=cv)
            self.featureR2[i] = np.mean(cvScore)
        return self

    def Transform(self, X, nSelFeatures = None):
        if nSelFeatures is None:
            self.evalFeatures()
            nSelFeatures = np.argmax(self.featureR2)+1
            
        selFeatures = self.featureRank[:nSelFeatures]
        return X[:, selFeatures]
                
# Feature selection with random test method
class RT(MCUVE):
    def fit(self):
        # calculate normal pls regression coefficient
        plsmodel0=PLSRegression(self.ncomp)
        plsmodel0.fit(self.x, self.y)
        # calculate noise reference regression coefficient
        plsCoef0=plsmodel0.coef_
        PLSCoef = np.zeros((self.nrep, self.x.shape[1]))
        for i in range(self.nrep):
            randomidx = list(range(self.x.shape[0]))
            np.random.shuffle(randomidx)
            ytrain = self.y[randomidx]
            plsModel = PLSRegression(self.ncomp)
            plsModel.fit(self.x, ytrain)
            PLSCoef[i, :] = plsModel.coef_.T
        plsCoef0 = np.tile(np.reshape(plsCoef0, [1, -1]), [ self.nrep, 1])
        criteria = np.sum(np.abs(PLSCoef) > np.abs(plsCoef0), axis=0)/self.nrep
        self.criteria = criteria
        self.featureRank = np.argsort(self.criteria)
        return self

    def evalFeatures(self, cv=5):
        self.featureR2 = []
        # Note: small P value indicating important feature
        
        for i in range(self.x.shape[1]):
            xi = self.x[:, self.featureRank[:i + 1]]
            if i<self.ncomp:
                regModel = LinearRegression()
            else:
                regModel = PLSRegression(min([self.ncomp, rank(xi)]))
            cvScore = cross_val_score(regModel, xi, self.y, cv=cv)
            self.featureR2[i] = np.mean(cvScore)
        return self


# Feature selection based on the criterion of C values
# Ref. [1]	Zhang J., et.al. A variable importance criterion for variable selection in near-infrared spectral analysis [J]. Science China-Chemistry, 2019, 62(2): 271-279.
class VC(RT):
    def fit(self, cv=5, isSmooth = True):
        # calculate normal pls regression coefficient
        nVar = self.x.shape[1]
        sampleMatrix = np.ndarray([self.nrep,self.x.shape[1]], dtype=int)
        sampleMatrix[:, :] = 0
        errVector = np.ndarray([self.nrep,1])
        # The number of variable in combination should less than the total variable number
        if nVar > self.ncomp:
            nSample = max([self.ncomp, nVar//10])
        else:
            nSample = max([1, nVar-1])
        sampleidx = range(self.x.shape[1])
        for i in range(self.nrep):
            sampleidx = shuffle(sampleidx)
            seli = sampleidx[:nSample]
            plsModel = PLSRegression(n_components=min([self.ncomp, rank(self.x[:, seli])]))
            plsModel.fit(self.x[:, seli], self.y)
            sampleMatrix[i, seli] = 1
            yhati=cross_val_predict(plsModel, self.x[:, seli], self.y, cv=cv)
            errVector[i] = np.sqrt(mean_squared_error(yhati, self.y))
        plsModel = PLSRegression(n_components=self.ncomp)
        plsModel.fit(sampleMatrix, errVector)
        self.criteria = plsModel.coef_.ravel()
        if self.nrep < 5*self.x.shape[0] and isSmooth:
           self.criteria =  smooth(polyorder=1).transform(self.criteria)
        self.featureRank = np.argsort(self.criteria)
        return self

# Recursive feature selection based on the criterion of C values
class MSVC:
    def __init__(self, x, y, ncomp=1, nrep=7000, ncut=50, testSize=0.2):
        self.x = x
        self.y = y
        # The number of latent components should not be larger than any dimension size of independent matrix
        self.ncomp = min([ncomp, rank(x)])
        self.nrep = nrep
        self.ncut = ncut
        self.testSize = testSize
        self.criteria = np.full([ncut, self.x.shape[1]], np.nan)
        self.featureR2 = np.empty(ncut)
        self.selFeature = None

    def calcCriteria(self):
        varidx = np.array(range(self.x.shape[1]))
        ncuti = np.logspace(np.log(self.x.shape[1]), np.log(1), self.ncut, base=np.e)
        ncuti = (np.round(ncuti)).astype(int)
        for i in range(self.ncut):
            vcModel = VC(self.x[:, varidx], self.y, self.ncomp, nrep=self.nrep)
            vcModel.fit(isSmooth = False)
            self.criteria[i, varidx] = vcModel.criteria
            var_ranked = np.argsort(vcModel.criteria)
            if i < self.ncut - 1:
                varidx = varidx[var_ranked[:ncuti[i+1]]]

    def evalCriteria(self, cv=3):
        for i in range(self.ncut):
            varSeli = ~np.isnan(self.criteria[i, :])
            xi = self.x[:,varSeli]
            if sum(varSeli) < self.ncomp:
                regModel = LinearRegression()
            else:
                regModel = PLSRegression(min([self.ncomp, rank(xi)]))
            cvScore = cross_val_score(regModel, xi, self.y, cv=cv)
            self.featureR2[i] = np.mean(cvScore)

    def cutFeature(self, *args):
        cuti = np.argmax(self.featureR2)
        self.selFeature = ~np.isnan(self.criteria[cuti, :])
        if len(args) != 0:
            returnx = list(args)
            i = 0
            for argi in args:
                if argi.shape[1] == self.x.shape[1]:
                    returnx[i] = argi[:, self.selFeature]
                i += 1
        return tuple(returnx)


def plotFeatureSelection(wv, X, selFeatures, methodNames = None,
                         ylabel = "Intensity (A.U.)", 
                         xlabel = "Wavelength (nm)",
                         title = "Feature selection results",
                         ax = None):
    scale = 0.4
    ofset = 1
    if ax is None:
        fig, ax = plt.subplots()
    
    ax.plot(wv,X.transpose())
    if isinstance(selFeatures[0], int) or \
        isinstance(selFeatures[0], np.int64):
        ax.eventplot(wv[selFeatures],
                     linelengths = scale*np.std(X),
                     lineoffsets = np.min(X)-(scale+ofset)*np.std(X))
        if methodNames is not None:
            ax.text(wv[0]-0.05*(wv[-1]-wv[0]),np.min(X)-(scale+ofset)*np.std(X),methodNames)
    else:
        for i in range(len(selFeatures)):
            ax.eventplot(wv[selFeatures[i]],
                         linelengths = scale*np.std(X),
                         lineoffsets = np.min(X)-(scale*2*i+ofset)*np.std(X))
            if methodNames is not None:
                ax.text(wv[0]-0.05*(wv[-1]-wv[0]), np.min(X)-(scale*2*i+ofset)*np.std(X),
                        methodNames[i])
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)


import numpy as np
import matplotlib.pyplot as plt

from pynir.utils import simulateNIR
from pynir.Calibration import plsda
from pynir.Calibration import multiClassificationReport,plot_confusion_matrix

from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix

# simulate NIR data
nclass = 4
X,y,wv = simulateNIR(nSample=200,nComp=10,refType=nclass, noise=1e-5)

Xtrain, Xtest, ytrain,ytest = train_test_split(X,y,test_size=0.2)

# estabilish PLS model
nComp = 10
plsdaModel = plsda(nComp = nComp)
plsdaModel.fit(Xtrain,ytrain)

# 10 fold cross validation for selecting optimal nlv
accuracy_cv = []
yhat_cv  = plsdaModel.crossValidation_predict(nfold = 10)
for i in range(nComp):
    report_cv = multiClassificationReport(ytrain, yhat_cv[:,i])
    accuracy_cv.append(np.mean([rep["accuracy"] for rep in report_cv.values()]))


fig,ax = plt.subplots()
ax.plot(np.arange(nComp)+1,accuracy_cv, marker = "*",label = "Accuracy$_c$$_v$")
ax.set_xlabel("nLV")
ax.set_ylabel("Accuracy")
ax.legend()
plt.show()


optLV = plsdaModel.get_optLV()  # optimized nLV based on cross validation

plsdaModel_opt = plsda(nComp = optLV)
plsdaModel_opt.fit(Xtrain,ytrain)
yhat_train_opt = plsdaModel_opt.predict(Xtrain)
yhat_test_opt = plsdaModel_opt.predict(Xtest)

cm_train_opt = confusion_matrix(y_true=ytrain, y_pred=yhat_train_opt)
cm_test_opt = confusion_matrix(y_true=ytest, y_pred=yhat_test_opt)



plot_confusion_matrix(cm_train_opt,np.unique(y),normalize=False,
                      title="Confusion matrix for prediction on training set")
plot_confusion_matrix(cm_test_opt,np.unique(y),normalize=False,
                      title="Confusion matrix for prediction on testing set")



import numpy as np
import matplotlib.pyplot as plt

from pynir.utils import simulateNIR

from pynir.OutlierDection import outlierDection_PLS

# simulate NIR data
X,y,wv = simulateNIR(nSample=200,nComp=10,noise=1e-5)

ODModel = outlierDection_PLS(ncomp=3)
Q, Tsq, Q_conf, Tsq_conf, idxOutlier = ODModel.fit(X, y).detect(X,y)
ODModel.plot_HotellingT2_Q(Q, Tsq, Q_conf, Tsq_conf)


import sys
from pathlib import Path

pynir_path = str(Path(__file__).parent.parent / 'src')

if pynir_path not in sys.path:
    sys.path.append(pynir_path)
    print("add pynir path")

from pynir.utils import simulateNIR_calibrationTransfer
from pynir.Calibration import pls, regresssionReport
from pynir.CalibrationTransfer import NS_PFCE,SS_PFCE,FS_PFCE,MT_PFCE
import matplotlib.pyplot as plt
import numpy as np

from scipy.io import loadmat

from sklearn.model_selection import train_test_split

import time


RawData = loadmat("./Data_Tablet.mat")
wv = RawData["wv"].ravel()
Xcal1 = RawData["Xcal1"]
Xcal2 = RawData["Xcal2"]
Xtrans1 = RawData["Xtrans1"]
Xtrans2 = RawData["Xtrans2"]
Xtest1 = RawData["Xtest1"]
Xtest2 = RawData["Xtest2"]
ycal = RawData["ycal"]
ytrans = RawData["ytrans"]
ytest = RawData["ytest"]


fig, ax = plt.subplots(2,sharex=True)
ax[0].plot(wv, np.transpose(Xcal1))
ax[0].set_title("First")

ax[1].plot(wv, np.transpose(Xcal2))
ax[1].set_title("Second")

fig.supxlabel("Wavelength (nm)")
fig.supylabel("Intensity(A.U.)")
plt.show()

nComp = 3
plsModel1 = pls(nComp=nComp).fit(Xcal1, ycal)

yhat1 = plsModel1.predict(Xtest1,nComp=nComp)
yhat2= plsModel1.predict(Xtest2,nComp=nComp)

fig, ax = plt.subplots(2,sharex=True,figsize=(8,16))
plsModel1.plot_prediction(ytest, yhat1,title = "First", ax = ax[0])
plsModel1.plot_prediction(ytest, yhat2, title= "Second", ax = ax[1])

fig, ax = plt.subplots()
ax.plot(wv, plsModel1.model['B'][1:,-1])
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Regression Coefficients")

## PFCE
thres = 0.98
constrType = 1

tic = time.time()
b1 = plsModel1.model['B'][:,-1]
NS_PFCE_model = NS_PFCE(thres=thres, constrType=constrType).fit(Xtrans1,Xtrans2,b1)
yhat2_NS_PFCE = NS_PFCE_model.transform(Xtest2)
plsModel1.plot_prediction(ytest, yhat2_NS_PFCE, title= "NS-PFCE")

fig, ax = plt.subplots()
ax.plot(wv, NS_PFCE_model.b2.x[1:])
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Regression Coefficients")
ax.set_title("NS-PFCE")
print("cost {:.2f} seconds".format(time.time()-tic))


tic = time.time()
b1 = plsModel1.model['B'][:,-1]
SS_PFCE_model = SS_PFCE(thres=thres, constrType=constrType).fit(Xtrans2,ytrans,b1)
yhat2_SS_PFCE = SS_PFCE_model.transform(Xtest2)
plsModel1.plot_prediction(ytest, yhat2_SS_PFCE, title= "SS-PFCE")

fig, ax = plt.subplots()
ax.plot(wv, SS_PFCE_model.b2.x[1:])
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Regression Coefficients")
ax.set_title("SS-PFCE")
print("cost {:.2f} seconds".format(time.time()-tic))


tic = time.time()
b1 = plsModel1.model['B'][:,-1]
FS_PFCE_model = FS_PFCE(thres=thres, constrType=constrType).fit(Xtrans1,Xtrans2,ytrans,b1)
yhat2_FS_PFCE = FS_PFCE_model.transform(Xtest2)
plsModel1.plot_prediction(ytest, yhat2_FS_PFCE, title= "FS-PFCE")

fig, ax = plt.subplots()
ax.plot(wv, FS_PFCE_model.b2.x[1:])
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Regression Coefficients")
ax.set_title("FS-PFCE")
print("cost {:.2f} seconds".format(time.time()-tic))


tic = time.time()
plsModel_global = pls(nComp=nComp).fit(np.vstack((Xcal1,Xtrans2)),
                                       np.vstack((ycal,ytrans)))
optLV_global = plsModel_global.get_optLV()
plsModel_global = pls(nComp=optLV_global).fit(np.vstack((Xcal1,Xtrans2)),
                                       np.vstack((ycal,ytrans)))

bglb = plsModel_global.model['B'][:,-1]
MT_PFCE_model = MT_PFCE(thres=thres, constrType=constrType)
MT_PFCE_model.fit([Xcal1,Xtrans2],(ycal,ytrans),bglb)
yhat1_MT_PFCE = MT_PFCE_model.transform(Xtest1,0)
yhat2_MT_PFCE = MT_PFCE_model.transform(Xtest2,1)

fig, ax = plt.subplots(2,sharex=True,figsize=(8,16))
plsModel1.plot_prediction(ytest, yhat1_MT_PFCE, title= "MT-PFCE_First", ax= ax[0])
plsModel1.plot_prediction(ytest, yhat2_MT_PFCE, title= "MT-PFCE_Second", ax= ax[1])

fig, ax = plt.subplots()
ax.plot(wv, MT_PFCE_model.B.x.reshape(2,-1).transpose()[1:,:])
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Regression Coefficients")
ax.set_title("MT-PFCE")
print("cost {:.2f} seconds".format(time.time()-tic))



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 11:01:16 2022

@author: chinn
"""
import numpy as np
from sklearn.linear_model import LinearRegression
from scipy import signal
import pywt

class snv():
    def __init__(self):
        pass
    
    def fit(self, X):
        self.mean = np.mean(X,axis=0)
        self.std = np.std(X,axis=0)
        return self
    
    def transform(self, X):
        return (X-self.mean[None,:])/self.std[None,:]
    
    def fit_transform(self, X):
        self.mean = np.mean(X,axis=0)
        self.std = np.std(X,axis=0)
        return self.transform(X)
    
class cwt():
    def __init__(self, wavelet = "morl", scale = 20):
        """
        Parameters
        ----------
        wavelet : string, optional
            Wavelet object or name:
            ['cgau1'-'cgau8','cmor','fbsp',
             'gaus1'-'gaus8','mexh','morl','shan'].
            For details about this wavelet, refer to https://pywavelets.readthedocs.io/en/latest/ref/cwt.html
            The default is "morl".
        scale : array_like, optional
            Wavelet scale to use. The default is 20.

        Returns
        -------
        None.

        """
        self.wavelet = wavelet
        self.scale = scale
    
    
    def getContinuousWavelet(self):
        return pywt.wavelist(kind = 'continuous')
    
    def getDiscreteWavelet(self):
        return pywt.wavelist(kind = 'discrete')
        
    def transform(self, X):
        nrow, ncol = X.shape
        
        Xcwt = pywt.cwt(X, self.scale, self.wavelet,axis=1)
        Xcwt = Xcwt[0].squeeze()

        return Xcwt

class msc():
    def __init__(self):
        pass
    
    def fit(self, X):
        self.mean = np.mean(X,axis=0)
        return self
        
    def transform(self, X):
        for i in range(X.shape[0]):
            X[i,:] = LinearRegression().fit(X[i,:][:,None], self.mean[:,None]).predict(X[i,:][:,None]).ravel()
        return X
        
    
    def fit_transform(self, X):
        return self.fit(X).transform(X)

class SG_filtering():
    def __init__(self,window_length = 13, polyorder=2, **kwargs):
        self.window_length = window_length
        self.polyorder = polyorder
        self.kwargs = kwargs
    
    def transform(self, X):
        return signal.savgol_filter(X, window_length = self.window_length, 
                             polyorder = self.polyorder, **self.kwargs)
    

    
class centralization():
    def __init__(self):
        pass
    
    def fit(self, X):
        self.mean = np.mean(X,axis=0)
        return self
    
    def transform(self, X):
        return X-self.mean[None,:]
    
    def fit_transform(self, X):
        return self.fit(X).transform(X)
    
# High level preprocessing function   
class derivate():
    def __init__(self,deriv = 1, window_length = 13, polyorder=2, **kwargs):
        self.deriv = deriv
        self.window_length = window_length
        self.polyorder = polyorder
        self.kwargs = kwargs
    
    def transform(self, X):
        return signal.savgol_filter(X, deriv=self.deriv, window_length = self.window_length, 
                             polyorder = self.polyorder,
                             delta = 1.0, **self.kwargs)
    
    
class smooth():
    def __init__(self,window_length = 13, polyorder=2, **kwargs):
        self.window_length = window_length
        self.polyorder = polyorder
        self.kwargs = kwargs
    
    def transform(self, X):
        return signal.savgol_filter(X,  window_length = self.window_length, 
                             polyorder = self.polyorder, deriv=0, **self.kwargs)


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 11:00:35 2022

@author: chinn
"""
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.cross_decomposition import PLSRegression
from sklearn.decomposition import TruncatedSVD

from scipy.optimize import (BFGS, SR1, Bounds,
                            NonlinearConstraint, minimize)


class PDS():
    def __init__(self, halfWindowSize = 7, regType = 'mlr',**kwargs):
        self.halfWindowSize = halfWindowSize
        self.regType = regType
        self.kwargs = kwargs

    def get_windowRange(halfWindowSize, nFeatrues, i):
        if i < halfWindowSize:
            band = np.arange(i)
        elif i > nFeatrues-halfWindowSize:
            band = np.arange(i)
        else:
            band = np.arange(i-halfWindowSize, i+halfWindowSize)
        return band

    def fit(self, X1, X2):
        """
        Parameters
        ----------
        X1 : 2D array
            Standard spectra of master.

        X2 : 2D array
            Standard spectra of master.
        """

        if X1.shape != X2.shape:
            raise("The dimension of two spectral matrix doesn't match.")

        Models = []

        for i in range(X1.shape[1]):
            if i < self.halfWindowSize:
                X2i = X2[:,:2*self.halfWindowSize+1]
            elif i >= self.halfWindowSize and i < X2.shape[1]-self.halfWindowSize:
                X2i = X2[:,i-self.halfWindowSize:i+self.halfWindowSize+1]
            elif i > X2.shape[1]-self.halfWindowSize:
                X2i = X2[:,X2.shape[1]-2*self.halfWindowSize-1:]



            if self.regType == 'mlr':
                Models.append(LinearRegression().fit(X2i,X1[:,i][:,None]))
            elif self.regType == 'pls':
                Models.append(PLSRegression(n_components=self.nComp).fit(X2i,X1[:,i][:,None]))

        self.Models = Models
        return self

    def transform(self, X):
        Xnew = np.zeros(X.shape)
        for i in range(X.shape[1]):
            if i < self.halfWindowSize:
                Xi = X[:,:2*self.halfWindowSize+1]
            elif i >= self.halfWindowSize and i < X.shape[1]-self.halfWindowSize:
                Xi = X[:,i-self.halfWindowSize:i+self.halfWindowSize+1]
            elif i > X.shape[1]-self.halfWindowSize:
                Xi = X[:,X.shape[1]-2*self.halfWindowSize-1:]

            Xnew[:,i] = self.Models[i].predict(Xi).ravel()
        return Xnew


class SST():
    def __init__(self, nComp=2):
        self.nComp = nComp

    def fit(self,X1,X2):
        """
        Parameters
        ----------
        X1 : 2D array
            Standard spectra of the first instrument.

        X2 : 2D array
            Standard spectra of the second instrument.
        """
        svd = TruncatedSVD(n_components=self.nComp).fit(np.hstack((X1,X2)))
        P = svd.components_
        nFeature_X1 = X1.shape[1]
        W1 = P[:,:nFeature_X1]
        W2 = P[:,nFeature_X1:]
        self.F = np.dot(np.linalg.pinv(W2),  (W1-W2))
        return self

    def transform(self,X):
        return X + np.dot(X,self.F)


class BS():
    """
    Ref Osborne B. G., Fearn T. Collaborative evaluation of universal
        calibrations for the measurement of protein and moisture in flour by near
        infrared reflectance [J]. International Journal of Food Science & Technology,
        1983, 18(4): 453-460.
    """
    def __init__(self):
        pass

    def fit(self, y1, y2):
        self.lrModel = LinearRegression().fit(y2.reshape(-1,1), y1.reshape(-1,1))
        return self

    def transform(self, y2):
        return self.lrModel.predict(y2.reshape(-1,1))


class NS_PFCE():
    def __init__(self,thres = 0.98, constrType = 1):
        self.thres = thres
        self.constrType = constrType

    def fit(self, X1, X2, b1):
        ntask = 2
        lb = -np.inf
        ub = 0.0
        if self.constrType == 1:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr1(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 2:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr2(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 3:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr3(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)

        b2 = minimize(lambda b2:cost_NS_PFCE(b1,b2,X1,X2),
                      x0 = b1, method='SLSQP', constraints=[nlc])
        self.b2 = b2
        return self

    def transform(self, X):
        return np.dot(np.hstack((np.ones([X.shape[0],1]), X)),
                      np.reshape(self.b2.x,(-1,1)))

class SS_PFCE():
    def __init__(self,thres = 0.98, constrType = 1):
        self.thres = thres
        self.constrType = constrType

    def fit(self, X2, y, b1):
        ntask = 2
        lb = -np.inf
        ub = 0.0
        if self.constrType == 1:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr1(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 2:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr2(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 3:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr3(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        else:
            raise("PFCE only support the constrint type of 1, 2, or 3")

        b2 = minimize(lambda b2:cost_SS_PFCE(b2,X2, np.ravel(y)),
                      x0 = b1, method='SLSQP', constraints=[nlc])
        self.b2 = b2
        return self

    def transform(self, X):
        return np.dot(np.hstack((np.ones([X.shape[0],1]), X)),
                      np.reshape(self.b2.x,(-1,1)))


class FS_PFCE():
    def __init__(self,thres = 0.98, constrType = 1):
        self.thres = thres
        self.constrType = constrType

    def fit(self, X1, X2, y, b1):
        ntask = 2
        lb = -np.inf
        ub = 0.0
        if self.constrType == 1:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr1(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 2:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr2(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)
        elif self.constrType == 3:
            nlc = NonlinearConstraint(
                lambda b2:pfce_constr3(np.hstack((b1.reshape(-1,1), b2.reshape(-1,1))),
                                       self.thres), lb, ub)

        b2 = minimize(lambda b2:cost_FS_PFCE(b1,b2,X1,X2,np.ravel(y)),
                      x0 = b1,  method='SLSQP', constraints=[nlc])
        self.b2 = b2
        return self

    def transform(self, X):
        return np.dot(np.hstack((np.ones([X.shape[0],1]), X)),
                      np.reshape(self.b2.x,(-1,1)))


class MT_PFCE():
    def __init__(self,thres = 0.98, constrType = 1):
        self.thres = thres
        self.constrType = constrType

    def fit(self, X, y, b1):
        ntask = len(X)
        y = [np.ravel(yi) for yi in y]
        self.ntask = ntask
        lb = -np.inf
        ub = 0.0
        if self.constrType == 1:
            nlc = NonlinearConstraint(
                lambda B:pfce_constr1(np.transpose(np.reshape(B,(ntask,-1))),
                                       self.thres), lb, ub)
        elif self.constrType == 2:
            nlc = NonlinearConstraint(
                lambda B:pfce_constr2(np.transpose(np.reshape(B,(ntask,-1))),
                                       self.thres), lb, ub)
        elif self.constrType == 3:
            nlc = NonlinearConstraint(
                lambda B:pfce_constr3(np.transpose(np.reshape(B,(ntask,-1))),
                                       self.thres), lb, ub)

        B = minimize(lambda B:cost_MT_PFCE(B.reshape(ntask,-1).transpose(),X,y),
                      x0 = np.tile(b1,ntask),method='SLSQP',constraints=[nlc])
        self.B = B
        return self

    def transform(self, X, itask):
        return np.dot(np.hstack((np.ones([X.shape[0],1]), X)),
                      ((self.B.x).reshape(self.ntask,-1).transpose()[:,itask]).reshape((-1,1)))


def cost_NS_PFCE(b1, b2, X1, X2):
    yhat1 = np.dot(np.hstack((np.ones([X1.shape[0],1]), X1)), b1)
    yhat2 = np.dot(np.hstack((np.ones([X2.shape[0],1]), X2)), b2)
    error = yhat1-yhat2
    return np.sum(error**2)


def cost_SS_PFCE(b2, X2, y):
    yhat2 = np.dot(np.hstack((np.ones([X2.shape[0],1]), X2)), b2)
    error = y-yhat2
    return np.sum(error**2)


def cost_FS_PFCE(b1, b2, X1, X2, y):
    yhat1 = np.dot(np.hstack((np.ones([X1.shape[0],1]), X1)), b1)
    yhat2 = np.dot(np.hstack((np.ones([X2.shape[0],1]), X2)), b2)
    error1 = yhat1-yhat2
    error2 = y-yhat2
    return np.sum(error1**2/len(error1)+error2**2/len(error2))


def cost_MT_PFCE(B, X, y):
    Error = []
    for i in range(B.shape[1]):
        yhati = np.dot(np.hstack((np.ones([X[i].shape[0],1]), X[i])), B[:,i])
        errori = y[i]-yhati
        Error.append(np.mean(errori**2))
    return np.sum(Error)


def pfce_constr1(B,thres):
    corr = np.corrcoef((B[1:,].transpose()))
    constr_cost  = thres - np.sum(np.triu(corr,1))/np.sum(np.arange(B.shape[1]))
    return constr_cost

def pfce_constr2(B,thres):
    B = B[1:,]
    Vars = np.diag(np.dot(B.transpose(),B))
    VarErr = 0
    for i in range(B.shape[1]):
        Err = B-B[:,i][:,None]
        Erri = Err;
        Erri = np.delete(Erri, i, axis =1)
        Varsi = Vars
        Varsi = np.delete(Varsi, i)
        VarErr += np.sum(Erri**2,axis = 0)/np.sqrt(Vars[i]*Varsi)
    dist = VarErr/(B.shape[1] * (B.shape[1]-1))
    constr_cost = thres - 1 + dist
    return constr_cost

def pfce_constr3(B,thres):
    B = B[1:,]
    norm1s = np.sum(np.abs(B),axis=0)
    norm1Err = 0
    for i in range(B.shape[1]):
        Err = B-B[:,i][:,None]
        Erri = Err
        Erri = np.delete(Erri, i,axis = 1)
        norm1si = norm1s
        norm1si = np.delete(norm1si, i)
        norm1Err += np.sum(np.abs(Erri),axis = 0)/np.sqrt(norm1s[i]*norm1si)
    dist = norm1Err/(B.shape[1] * (B.shape[1]-1))
    constr_cost = thres - 1 + dist
    return constr_cost



import numpy as np
import matplotlib.pyplot as plt
from pynir.utils import simulateNIR

X,y,wv = simulateNIR(nSample=200,nComp=10,noise=1e-5)

fig, ax = plt.subplots()
ax.plot(wv,np.transpose(X))
ax.set_xlabel("wavelength (nm)")
ax.set_ylabel("Intesntiy (a.u.)")
plt.show()

fig, ax = plt.subplots()
ax.hist(y)
ax.set_xlabel("Reference values")
ax.set_ylabel("Count")
plt.show()


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 11:00:35 2022

@author: chinn
"""
import numpy as np
from sklearn.model_selection import train_test_split, KFold

from sklearn.metrics import mean_squared_error, r2_score
from sklearn.metrics import accuracy_score, recall_score, f1_score, confusion_matrix

from sklearn.decomposition import PCA
from scipy.spatial.distance import cdist

from sklearn.cross_decomposition import PLSRegression
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.svm import LinearSVC, SVC
from sklearn.ensemble import RandomForestClassifier

from sklearn.model_selection import GridSearchCV, cross_val_predict

import matplotlib.pyplot as plt


class nirData():
    def __init__(self, X, y, trainIdx=None, testIdx=None):
        self.X = X
        self.y = y
        self.trainIdx = trainIdx
        self.testIdx = testIdx

    def get_train_test_Data(self):
        if self.trainIdx == None or self.testIdx == None:
            sampleSplit_random()
        return (self.X[self.trainIdx,:],self.y[self.trainIdx,:],
                self.X[self.testIdx,:],self.y[self.testIdx,:])



class classification():
    def __init__(self, X, y, method = None, **kwargs):
        self.X = X
        self.y = y
        self.method = method


    def train(self):
        pass

    def predict():
        pass

class regression():
    def __init__(self, X, y, method = None, **kwargs):
        self.X = X
        self.y = y
        self.method = method


    def train(self):
        pass

    def predict():
        pass


class pls:
    def __init__(self, nComp = 2):
        self.nComp = nComp

    def fit(self, X, y):
        self.X = X
        self.y = y
        meanX = np.mean(X,axis = 0)
        meany = np.mean(y)
        Xcentered = X - meanX
        ycentered = y - meany
        model = simpls(Xcentered, ycentered, self.nComp)
        meanX_hat = -1 * np.dot(meanX, model['B']) + meany
        model['B'] = np.append(meanX_hat[np.newaxis,:], model['B'],axis=0)
        self.model = model
        return self

    def predict(self, Xnew, nComp = None):
        if nComp is None:
            B = self.model['B'][:,-1]
        else:
            B = self.model['B'][:, nComp-1]
        if Xnew.shape[1] != B.shape[0]-1:
            raise ValueError('The feature number of predictor is isconsistent with that of indepnentent.')
        Xnew = np.append(np.ones([Xnew.shape[0],1]), Xnew, axis=1)
        ynew_hat = np.dot(Xnew,B)
        return ynew_hat

    def crossValidation_predict(self, nfold = 10):
        X = self.X
        y = self.y
        yhat = np.zeros((y.shape[0],self.nComp))
        model = pls(nComp = self.nComp)
        for train, test in KFold(n_splits=nfold).split(X):
            model.fit(X[train,:], y[train])
            yhat[test,:] = model.predict(X[test,:],np.arange(self.nComp)+1)
        return yhat

    def get_optLV(self, nfold = 10):
        yhat_cv = self.crossValidation_predict(nfold)
        rmsecv = []
        r2cv = []
        for i in range(yhat_cv.shape[1]):
            reportcv = regresssionReport(self.y, yhat_cv[:,i])
            rmsecv.append(reportcv["rmse"])
            r2cv.append(reportcv["r2"])
        optLV = int(np.argmin(rmsecv)+1)
        self.optLV = optLV
        return optLV

    def transform(self, Xnew):
        meanX = np.mean(self.X, axis = 0)
        Xnew_c = Xnew - meanX
        Tnew = np.dot(Xnew_c,self.model['x_weights'])
        return Tnew

    def calc_VIP(self, nComp = None):
        if nComp==None:
            nComp=self.model['B'].shape[1]
        b = self.model['B'][1:,nComp-1]
        T = self.model['x_scores']
        W = self.model['x_weights']
        up=0
        low=0
        for i in range(T.shape[1]):
            up += b**2 * np.dot(T[:,i],T[:,i]) * ((W[:,i])/np.linalg.norm(W[:,i]))**2
            low += b**2 * np.dot(T[:,i],T[:,i])
        VIP = np.sqrt(len(b)*up/low)
        return VIP

    def plot_prediction(self, y, yhat, xlabel = "Reference", ylabel = "Prediction", title = "", ax = None):
        report = regresssionReport(y,yhat)
        if ax == None:
            fig, ax = plt.subplots()
        ax.plot([np.min(y)*0.95,np.max(y)*1.05],[np.min(y)*0.95,np.max(y)*1.05],
                color = 'black',label = "y=x")
        ax.scatter(y, yhat,color = 'tab:green', marker='*', label ='Prediction')
        ax.text(0.7, 0.03,
                "RMSEP = {:.4f}\nR$^2$ = {:.2}".format(report["rmse"],report["r2"]),
                transform = ax.transAxes)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend(loc='upper left', bbox_to_anchor=(0.0, 1.0))
        ax.set_title(title)


class plsda(PLSRegression):
    def __init__(self, n_components=2, scale=True, **kwargs):
        super().__init__(n_components=n_components, scale=scale, **kwargs)
        self.lda = LinearDiscriminantAnalysis()

    def fit(self, X, y):
        self.X = X
        self.y = y
        super().fit(X, y)
        self.lda.fit(self.x_scores_, y)
        return self

    def predict(self, X):
        return self.lda.predict(self.transform(X))

    def predict_log_proba(self, X):
        return self.lda.predict_log_proba(self.predict(X))

    def predict_proba(self, X):
        return self.lda.predict_proba(self.predict(X))

    def crossValidation_predict(self, nfold = 10):
        X = self.X
        y = self.y
        yhat = np.zeros((y.shape[0],self.n_components))
        for i in range(self.n_components):
            model = plsda(n_components = i+1)
            for train, test in KFold(n_splits=nfold).split(X):
                model.fit(X[train,:], y[train])
                yhat[test,i] = model.predict(X[test,:])
        return yhat

    def get_optLV(self, nfold = 10):
        yhat_cv = self.crossValidation_predict(nfold)
        accuracy_cv = []
        for i in range(yhat_cv.shape[1]):
            if len(self.lda.classes_) == 2 :
                report_cv = binaryClassificationReport(self.y, yhat_cv[:,i])
                accuracy_cv.append(report_cv["accuracy"])
            elif len(self.lda.classes_) > 2:
                report_cv = multiClassificationReport(self.y, yhat_cv[:,i])
                accuracy_tmp = [rep["accuracy"] for rep in report_cv.values()]
                accuracy_cv.append(sum(accuracy_tmp))

        optLV = int(np.argmax(accuracy_cv)+1)
        self.optLV = optLV
        return optLV

    def get_confusion_matrix(self, X, y):
        yhat = self.predict(X)
        cm = confusion_matrix(y, yhat)
        return cm

    def get_vip(self):
        t = self.x_scores_ / self.x_scores_.std(axis=0)
        s = np.sum(t**2, axis=1)
        s /= np.sum(t**2)
        vip = s * np.sum(t**2, axis=1)
        return vip

    def permutation_test(self, X, y, n_repeats=100,n_jobs = None):
        # Initialize arrays to store Q2 and R2 values
        q2 = np.zeros(n_repeats)
        r2 = np.zeros(n_repeats)
        permutation_ratio = np.zeros(n_repeats)
        # Perform the permutation test
        for i in range(n_repeats):
            # Shuffle the target variable
            y_shuffled = np.random.permutation(y)

            # Fit the model to the shuffled target variable
            self.fit(X, y_shuffled)

            # Calculate the cross-validated Q2 and R2 values
            y_pred = cross_val_predict(self, X, y_shuffled, cv=10, n_jobs = n_jobs)
            q2[i] = self.score(X, y_shuffled)
            r2[i] = r2_score(y_shuffled, y_pred)
            permutation_ratio[i] = np.sum(y_shuffled != y) / len(y)
        return q2, r2, permutation_ratio

class lsvc(LinearSVC):# linear svc
    def get_optParams(self, X, y, Params = None, nfold = 10, n_jobs = None):
        if Params is None:
            Params = {'C': np.logspace(-4, 5, 10),
                      'penalty': ('l1', 'l2')}
        self.gsh = GridSearchCV(estimator=self,  param_grid=Params,
                           cv = nfold, n_jobs = n_jobs)
        self.gsh.fit(X, y)
        return self.gsh.best_params_

    def get_confusion_matrix(self, X, y):
        yhat = self.predict(X)
        cm = confusion_matrix(y, yhat)
        return cm

class svc(SVC):# linear svc
    def get_optParams(self, X, y, Params = None, nfold = 10, n_jobs = None):
        if Params is None:
            Params = {'C': np.logspace(-4, 5, 10),
                      'gamma':np.logspace(-4, 5, 10),
                      'kernel': ('poly', 'rbf', 'sigmoid')}
        self.gsh = GridSearchCV(estimator=self,  param_grid=Params,
                           cv = nfold, n_jobs = n_jobs)
        self.gsh.fit(X, y)
        return self.gsh.best_params_

    def get_confusion_matrix(self, X, y):
        yhat = self.predict(X)
        cm = confusion_matrix(y, yhat)
        return cm

class rf(RandomForestClassifier):
    def get_optParams(self, X, y, Params = None, nfold = 10, n_jobs = None):
        if Params is None:
            Params = {'n_estimators': np.arange(100)+1,
                      'max_depth': np.arange(3)+1}
        self.gsh = GridSearchCV(estimator=self,  param_grid=Params,
                           cv = nfold, n_jobs = n_jobs)
        self.gsh.fit(X, y)
        return self.gsh.best_params_

    def get_confusion_matrix(self, X, y):
        yhat = self.predict(X)
        cm = confusion_matrix(y, yhat)
        return cm


class multiClass_to_binaryMatrix():
    def __init__(self):
        pass

    def fit(self, x):
        self.classes = np.unique(x)
        return self

    def transform(self, x):
        Xnew = np.zeros((len(x),len(self.classes)), dtype=int)
        if len(self.classes) > 2 :
            for i, classi in enumerate(self.classes):
                Xnew[:,i] = x==classi
        return Xnew

    def reTransform(self, xnew):
        x = [np.classes(np.where(xnew[i,:])) for i in range(xnew.shape[0])]
        return x

def plot_confusion_matrix(cm,
                          target_names,
                          title='Confusion matrix',
                          cmap=None,
                          normalize=True):
    """
    given a sklearn confusion matrix (cm), make a nice plot

    Arguments
    ---------
    cm:           confusion matrix from sklearn.metrics.confusion_matrix

    target_names: given classification classes such as [0, 1, 2]
                  the class names, for example: ['high', 'medium', 'low']

    title:        the text to display at the top of the matrix

    cmap:         the gradient of the values displayed from matplotlib.pyplot.cm
                  see http://matplotlib.org/examples/color/colormaps_reference.html
                  plt.get_cmap('jet') or plt.cm.Blues

    normalize:    If False, plot the raw numbers
                  If True, plot the proportions

    Usage
    -----
    plot_confusion_matrix(cm           = cm,                  # confusion matrix created by
                                                              # sklearn.metrics.confusion_matrix
                          normalize    = True,                # show proportions
                          target_names = y_labels_vals,       # list of names of the classes
                          title        = best_estimator_name) # title of graph

    Citiation
    ---------
    http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html

    """

    accuracy = np.trace(cm) / np.sum(cm).astype('float')
    misclass = 1 - accuracy

    if cmap is None:
        cmap = plt.get_cmap('Blues')

    plt.figure(figsize=(8, 6))
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    if target_names is not None:
        tick_marks = np.arange(len(target_names))
        plt.xticks(tick_marks, target_names, rotation=45)
        plt.yticks(tick_marks, target_names)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]


    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            if normalize:
                plt.text(j, i, "{:0.4f}".format(cm[i, j]),
                         horizontalalignment="center",
                         color="white" if cm[i, j] > thresh else "black")
            else:
                plt.text(j, i, "{:,}".format(cm[i, j]),
                         horizontalalignment="center",
                         color="white" if cm[i, j] > thresh else "black")


    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(accuracy, misclass))
    plt.show()

def multiClassificationReport(ytrue,ypred):
    labels = np.unique(ytrue)
    report = dict()
    for labeli in labels:
        report[labeli] = binaryClassificationReport(ytrue = ytrue == labeli, ypred = ypred == labeli)
    return report

def binaryClassificationReport(ytrue,ypred):
    if len(np.unique(ytrue))>2:
        raise("Use the multiClassificationReport function for multiple classification.")
    else:
        tn, fp, fn, tp = confusion_matrix(ytrue,ypred).ravel()
        report = dict()
        report["accuracy"] = accuracy_score(ytrue, ypred)
        report["sensitivity"] = recall_score(ytrue, ypred)#recall
        report["specificity"] = tn/(tn+fp)
        report["f1"] = f1_score(ytrue, ypred)
        return report

def regresssionReport(ytrue,ypred):
    report = dict()
    report["rmse"] = mean_squared_error(ytrue, ypred, squared=False)
    report["r2"] =  r2_score(ytrue, ypred)
    return report

def simpls(X, y, nComp):
    '''
    Partial Least Squares, SIMPLS
    Ref https://github.com/freesiemens/SpectralMultivariateCalibration/blob/master/pypls.py
    :param X:
    :param y:
    :param nComp:
    :return:
    '''
    n_samples, n_variables = X.shape
    if np.ndim(y) == 1:
        y = y[:, np.newaxis]
    if n_samples != y.shape[0]:
        raise ValueError('The number of independent and dependent variable are inconsistent')

    nComp = np.min((nComp, n_samples, n_variables))
    V = np.zeros((n_variables, nComp))
    x_scores = np.zeros((n_samples, nComp))  # X scores (standardized)
    x_weights = np.zeros((n_variables, nComp))  # X weights
    x_loadings = np.zeros((n_variables, nComp))  # X loadings
    y_loadings = np.zeros((1, nComp))  # Y loadings
    y_scores = np.zeros((n_samples, nComp))  # Y scores
    s = np.dot(X.T, y).ravel()  # cross-product matrix between the X and y_data
    for i in range(nComp):
        r = s
        t = np.dot(X, r)
        tt = np.linalg.norm(t)
        t = t / tt
        r = r / tt
        p = np.dot(X.T, t)
        q = np.dot(y.T, t)
        u = np.dot(y, q)
        v = p  # P的正交基
        if i > 0:
            v = v - np.dot(V, np.dot(V.T, p))  # Gram-Schimidt orthogonal
            u = u - np.dot(x_scores, np.dot(x_scores.T, u))
        v = v / np.linalg.norm(v)
        s = s - np.dot(v, np.dot(v.T, s))
        x_weights[:, i] = r
        x_scores[:, i] = t
        x_loadings[:, i] = p
        y_loadings[:, i] = q
        y_scores[:, i] = u
        V[:, i] = v
    B = np.cumsum(np.dot(x_weights, np.diag(y_loadings.ravel())), axis=1)
    return {'B': B, 'x_scores': x_scores, 'x_loadings': x_loadings, 'y_loadings': y_loadings, \
            'x_scores_weights': x_weights, 'x_weights': x_weights, 'y_scores':y_scores}



def sampleSplit_random(X,test_size=0.25, random_state=1, shuffle=False):
    sampleIdx = np.arange(X.shape[0])
    trainIdx, testIdx = train_test_split(sampleIdx,test_size=test_size,
                                         random_state=random_state,
                                         shuffle=shuffle)
    return trainIdx, testIdx

def sampleSplit_KS(X, test_size=0.25, metric='euclidean', *args, **kwargs):
    """Kennard Stone Sample Split method
    Parameters
    ----------
    spectra: ndarray, shape of i x j
        i spectrums and j variables (wavelength/wavenumber/ramam shift and so on)
    test_size : float, int
        if float, then round(i x (1-test_size)) spectrums are selected as test data, by default 0.25
        if int, then test_size is directly used as test data size
    metric : str, optional
        The distance metric to use, by default 'euclidean'
        See scipy.spatial.distance.cdist for more infomation
    Returns
    -------
    select_pts: list
        index of selected spetrums as train data, index is zero based
    remaining_pts: list
        index of remaining spectrums as test data, index is zero based
    References
    --------
    Kennard, R. W., & Stone, L. A. (1969). Computer aided design of experiments.
    Technometrics, 11(1), 137-148. (https://www.jstor.org/stable/1266770)
    """
    Xscore = PCA(n_components=2).fit_transform(X)
    distance = cdist(Xscore, Xscore, metric=metric, *args, **kwargs)
    select_pts = []
    remaining_pts = [x for x in range(distance.shape[0])]

    # first select 2 farthest points
    first_2pts = np.unravel_index(np.argmax(distance), distance.shape)
    select_pts.append(first_2pts[0])
    select_pts.append(first_2pts[1])

    # remove the first 2 points from the remaining list
    remaining_pts.remove(first_2pts[0])
    remaining_pts.remove(first_2pts[1])

    for i in range(round(X.shape[0]*(1-test_size)) - 2):
        # find the maximum minimum distance
        select_distance = distance[select_pts, :]
        min_distance = select_distance[:, remaining_pts]
        min_distance = np.min(min_distance, axis=0)
        max_min_distance = np.max(min_distance)

        # select the first point (in case that several distances are the same, choose the first one)
        points = np.argwhere(select_distance == max_min_distance)[:, 1].tolist()
        for point in points:
            if point in select_pts:
                pass
            else:
                select_pts.append(point)
                remaining_pts.remove(point)
                break
    return select_pts, remaining_pts



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019.4.30
# @Author  : FrankEl
# @File    : Feature_selection_demo_mcuve.py

import matplotlib.pyplot as plt
import numpy as np

from pynir.utils import simulateNIR
from pynir.Calibration import pls
from pynir.FeatureSelection import MCUVE,RT,VC,plotFeatureSelection

# simulate NIR data
X,y,wv = simulateNIR(nSample=200,nComp=10,noise=1e-5)



# estabilish PLS model
nComp = 10
plsModel = pls(nComp = nComp)
plsModel.fit(X,y)

# 10 fold cross validation for selecting optimal nComp
yhat = plsModel.predict(X,nComp=7)

fig, ax = plt.subplots(2)
ax[0].plot(wv, plsModel.model["B"][1:,-1])
ax[0].set_xlabel("Wavelength")
ax[0].set_ylabel("Intensity")
ax[0].set_title("Regression Coefficients")
plsModel.plot_prediction(y, yhat,ax = ax[1])


nSel = 100
nrep = 500
# Feature selection
vip = plsModel.calc_VIP()
featureSelected_vip = np.where(vip>1)

mcModel = MCUVE(X, y, nComp, nrep=nrep).fit()
featureSelected_MC_UVE = mcModel.featureRank[:nSel]

rtModel = RT(X, y, nComp, nrep=nrep).fit()
featureSelected_RT = rtModel.featureRank[:nSel]

vcModel = VC(X, y, nComp, nrep=nrep).fit()
featureSelected_VC = vcModel.featureRank[:nSel]



plotFeatureSelection(wv, X,
                     [featureSelected_vip,
                         featureSelected_MC_UVE,
                         featureSelected_RT,
                         featureSelected_VC], 
                     methodNames=["VIP","MC-UVE","RT","VC"])

fig, ax = plt.subplots()
mc = mcModel.criteria
rt = rtModel.criteria
vc = vcModel.criteria
vip = (vip - min(vip))/(max(vip)-min(vip))
mc = (mc-min(mc))/(max(mc)-min(mc))
rt = (rt-min(rt))/(max(rt)-min(rt))
vc = (vc-min(vc))/(max(vc)-min(vc))
ax.plot(wv,vip, label ='VIP')
ax.plot(wv,mc, label ='MC-UVE')
ax.plot(wv,rt, label ='RT')
ax.plot(wv,vc, label ='VC')
plt.legend()


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 26 10:42:15 2023

@author: chinn

"""
import numpy as np
from pynir.CalibrationTransfer import cost_MT_PFCE


B = np.array([[10,2,3,4],[11,2.1,3.1,4.1]])
B = B.transpose()
X = [np.array([[2,3,4],[2,3,4]]),np.array([[0.2,0.3,0.4],[0.2,0.3,0.4]])]
y = [np.array([39.2]),np.array([14.49]),]
c = cost_MT_PFCE(B, X, y)

from pynir.CalibrationTransfer import pfce_constr1, pfce_constr2,pfce_constr3
c1 = pfce_constr1(B,0.98)
c2 = pfce_constr2(B,0.98)
c3 = pfce_constr3(B,0.98)
